@@ -15,6 +15,20 @@
         text_fuzzy->n_mallocs++;                                \
     }
 
+typedef enum {
+   tfp_ok,
+   tfp_unicode_failure,
+   tfp_text_fuzzy_error,
+}
+tfp_status_t;
+
+#define TFPCALL(x) {				\
+	tfp_status_t status = x;		\
+	if (status != tfp_ok) {			\
+	    return status;			\
+	}					\
+    }
+
 /* Send a bad return value from one of the C routines in
    "text-fuzzy.c" back to the user via Perl's error handlers. The
    parameters "file_name" and "line_number" are the name of the C file
@@ -60,7 +74,8 @@ static void fake_length (text_fuzzy_t * text_fuzzy, int minimum)
 
 /* Allocate the memory for b. */
 
-static void allocate_b_unicode (text_fuzzy_t * text_fuzzy, int b_length)
+static void 
+allocate_b_unicode (text_fuzzy_t * text_fuzzy, int b_length)
 {
 
     if (! text_fuzzy->b.unicode) {
@@ -87,34 +102,30 @@ static void allocate_b_unicode (text_fuzzy_t * text_fuzzy, int b_length)
    characters, use Perl's Unicode handlers to turn it into a string of
    integers. */
 
-static void sv_to_int_ptr (SV * text, text_fuzzy_string_t * tfs)
+static tfp_status_t
+sv_to_int_ptr (SV * text, text_fuzzy_string_t * tfs)
 {
     int i;
     const U8 * utf;
-    STRLEN curlen;
+    const U8 * send;
     STRLEN length;
-    const unsigned char * stuff;
 
-    stuff = (const unsigned char *) SvPV (text, length);
-
-    utf = stuff;
-    curlen = length;
+    utf = (const U8 *) SvPV (text, length);
+    send = utf + length;
     for (i = 0; i < tfs->ulength; i++) {
         STRLEN len;
-
-	/* The documentation for "utf8n_to_uvuni" can be found in
-	   "perldoc perlapi". There is an online version here:
-	   "http://perldoc.perl.org/perlapi.html#Unicode-Support". */
-
-        tfs->unicode[i] = utf8n_to_uvuni (utf, curlen, & len, 0);
-        curlen -= len;
+        tfs->unicode[i] = (int) utf8_to_uvchr_buf (utf, send, & len);
+	if (len == -1 || tfs->unicode[i] == 0) {
+	    return tfp_unicode_failure;
+	}
         utf += len;
     }
+    return tfp_ok;
 }
 
 /* Convert a Perl SV into the text_fuzzy_t structure. */
 
-static void
+static tfp_status_t
 sv_to_text_fuzzy (SV * text, text_fuzzy_t ** text_fuzzy_ptr)
 {
     STRLEN length;
@@ -122,6 +133,7 @@ sv_to_text_fuzzy (SV * text, text_fuzzy_t ** text_fuzzy_ptr)
     text_fuzzy_t * text_fuzzy;
     int i;
     int is_utf8;
+    char * copy;
 
     /* Allocate memory for "text_fuzzy". */
     get_memory (text_fuzzy, 1, text_fuzzy_t);
@@ -129,12 +141,13 @@ sv_to_text_fuzzy (SV * text, text_fuzzy_t ** text_fuzzy_ptr)
 
     /* Copy the string in "text" into "text_fuzzy". */
     stuff = (const unsigned char *) SvPV (text, length);
-    text_fuzzy->text.length = length;
-    get_memory (text_fuzzy->text.text, length + 1, char);
+    get_memory (copy, length + 1, char);
     for (i = 0; i < (int) length; i++) {
-        text_fuzzy->text.text[i] = stuff[i];
+        copy[i] = stuff[i];
     }
-    text_fuzzy->text.text[text_fuzzy->text.length] = '\0';
+    copy[length] = '\0';
+    text_fuzzy->text.length = length;
+    text_fuzzy->text.text = copy;
     is_utf8 = SvUTF8 (text);
     if (is_utf8) {
 
@@ -146,7 +159,7 @@ sv_to_text_fuzzy (SV * text, text_fuzzy_t ** text_fuzzy_ptr)
 
 	get_memory (text_fuzzy->text.unicode, text_fuzzy->text.ulength, int);
 
-	sv_to_int_ptr (text, & text_fuzzy->text);
+	TFPCALL(sv_to_int_ptr(text, & text_fuzzy->text));
 
 	if (text_fuzzy->text.ulength > 0) {
 	    /* Generate the Unicode alphabet. */
@@ -159,13 +172,16 @@ sv_to_text_fuzzy (SV * text, text_fuzzy_t ** text_fuzzy_ptr)
 	    TEXT_FUZZY (generate_alphabet (text_fuzzy));
 	}
     }
+    TEXT_FUZZY (allocate_edits (text_fuzzy));
     * text_fuzzy_ptr = text_fuzzy;
+    return tfp_ok;
 }
 
-static void
+static tfp_status_t
 sv_to_text_fuzzy_string (SV * word, text_fuzzy_t * text_fuzzy)
 {
     STRLEN length;
+    char * nonu;
     text_fuzzy->b.text = SvPV (word, length);
     text_fuzzy->b.allocated = 0;
     text_fuzzy->b.length = length;
@@ -184,22 +200,24 @@ sv_to_text_fuzzy_string (SV * word, text_fuzzy_t * text_fuzzy)
 
 	    text_fuzzy->b.length = text_fuzzy->b.ulength;
 	    text_fuzzy->b.allocated = 1;
-	    get_memory (text_fuzzy->b.text, text_fuzzy->b.length + 1, char);
+	    get_memory (nonu, text_fuzzy->b.length + 1, char);
 	    for (i = 0; i < text_fuzzy->b.ulength; i++) {
 		int c;
 
 		c = text_fuzzy->b.unicode[i];
 		if (c <= 0x80) {
-		    text_fuzzy->b.text[i] = c;
+		    nonu[i] = c;
 		}
 		else {
 		    /* Put a non-matching character in there. */
 
-		    text_fuzzy->b.text[i] = text_fuzzy->invalid_char;
+		    nonu[i] = text_fuzzy->invalid_char;
 		}
 	    }
+	    text_fuzzy->b.text = nonu;
 	}
     }
+    return tfp_ok;
 }
 
 static void
